/*
  This file is part of TALER
  Copyright (C) 2023, 2024, 2025, 2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_get-aml-OFFICER_PUB-kyc-statistics-NAMES.c
 * @brief Implementation of the /aml/$OFFICER_PUB/kyc-statistics/$NAME request
 * @author Christian Grothoff
 */
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include "taler/taler_json_lib.h"
#include "exchange_api_handle.h"
#include "taler/taler_signatures.h"
#include "exchange_api_curl_defaults.h"
#include "taler/exchange/get-aml-OFFICER_PUB-kyc-statistics-NAMES.h"


/**
 * @brief A GET /aml/$OFFICER_PUB/kyc-statistics/$NAMES Handle (new API)
 */
struct TALER_EXCHANGE_GetAmlKycStatisticsHandle
{

  /**
   * The base URL of the exchange.
   */
  char *base_url;

  /**
   * The full URL for this request.
   */
  char *url;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_EXCHANGE_GetAmlKycStatisticsCallback cb;

  /**
   * Closure for @e cb.
   */
  TALER_EXCHANGE_GET_AML_KYC_STATISTICS_RESULT_CLOSURE *cb_cls;

  /**
   * CURL context to use.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Public key of the AML officer.
   */
  struct TALER_AmlOfficerPublicKeyP officer_pub;

  /**
   * Private key of the AML officer.
   */
  struct TALER_AmlOfficerPrivateKeyP officer_priv;

  /**
   * Space-separated list of event type names to count.
   */
  char *names;

  /**
   * Options for this request.
   */
  struct
  {
    /**
     * Start date for statistics window. Zero means "from the beginning".
     */
    struct GNUNET_TIME_Timestamp start_date;

    /**
     * End date for statistics window. #GNUNET_TIME_UNIT_FOREVER_ABS means "up to now".
     */
    struct GNUNET_TIME_Timestamp end_date;
  } options;

};


/**
 * Parse the provided statistics data from the "200 OK" response.
 *
 * @param[in,out] aksh handle (callback may be zero'ed out)
 * @param json json reply
 * @return #GNUNET_OK on success, #GNUNET_SYSERR on error
 */
static enum GNUNET_GenericReturnValue
parse_stats_ok_new (
  struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *aksh,
  const json_t *json)
{
  struct TALER_EXCHANGE_GetAmlKycStatisticsResponse lr = {
    .hr.reply = json,
    .hr.http_status = MHD_HTTP_OK
  };
  const json_t *jstatistics;
  struct GNUNET_JSON_Specification spec[] = {
    GNUNET_JSON_spec_array_const ("statistics",
                                  &jstatistics),
    GNUNET_JSON_spec_end ()
  };

  if (GNUNET_OK !=
      GNUNET_JSON_parse (json,
                         spec,
                         NULL,
                         NULL))
  {
    GNUNET_break_op (0);
    return GNUNET_SYSERR;
  }
  lr.details.ok.statistics_length = json_array_size (jstatistics);
  {
    struct TALER_EXCHANGE_GetAmlKycStatisticsEventCounter statistics[
      GNUNET_NZL (lr.details.ok.statistics_length)];
    json_t *obj;
    size_t idx;

    memset (statistics,
            0,
            sizeof (statistics));
    lr.details.ok.statistics = statistics;
    json_array_foreach (jstatistics, idx, obj)
    {
      struct TALER_EXCHANGE_GetAmlKycStatisticsEventCounter *ec
        = &statistics[idx];
      struct GNUNET_JSON_Specification ispec[] = {
        GNUNET_JSON_spec_string ("event",
                                 &ec->name),
        GNUNET_JSON_spec_uint64 ("counter",
                                 &ec->counter),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (obj,
                             ispec,
                             NULL,
                             NULL))
      {
        GNUNET_break_op (0);
        return GNUNET_SYSERR;
      }
    }
    aksh->cb (aksh->cb_cls,
              &lr);
    aksh->cb = NULL;
  }
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP GET /aml/$OFFICER_PUB/kyc-statistics/$NAMES request.
 *
 * @param cls the `struct TALER_EXCHANGE_GetAmlKycStatisticsHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response parsed JSON result, NULL on error
 */
static void
handle_get_aml_kyc_statistics_finished (void *cls,
                                        long response_code,
                                        const void *response)
{
  struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *aksh = cls;
  const json_t *j = response;
  struct TALER_EXCHANGE_GetAmlKycStatisticsResponse lr = {
    .hr.reply = j,
    .hr.http_status = (unsigned int) response_code
  };

  aksh->job = NULL;
  switch (response_code)
  {
  case 0:
    lr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    if (GNUNET_OK !=
        parse_stats_ok_new (aksh,
                            j))
    {
      GNUNET_break_op (0);
      lr.hr.http_status = 0;
      lr.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
      break;
    }
    GNUNET_assert (NULL == aksh->cb);
    TALER_EXCHANGE_get_aml_kyc_statistics_cancel (aksh);
    return;
  case MHD_HTTP_NO_CONTENT:
    break;
  case MHD_HTTP_BAD_REQUEST:
    lr.hr.ec = TALER_JSON_get_error_code (j);
    lr.hr.hint = TALER_JSON_get_error_hint (j);
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    break;
  case MHD_HTTP_FORBIDDEN:
    lr.hr.ec = TALER_JSON_get_error_code (j);
    lr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_NOT_FOUND:
    lr.hr.ec = TALER_JSON_get_error_code (j);
    lr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_URI_TOO_LONG:
    lr.hr.ec = TALER_JSON_get_error_code (j);
    lr.hr.hint = TALER_JSON_get_error_hint (j);
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    lr.hr.ec = TALER_JSON_get_error_code (j);
    lr.hr.hint = TALER_JSON_get_error_hint (j);
    /* Server had an internal issue; we should retry, but this API
       leaves this to the application */
    break;
  default:
    /* unexpected response code */
    GNUNET_break_op (0);
    lr.hr.ec = TALER_JSON_get_error_code (j);
    lr.hr.hint = TALER_JSON_get_error_hint (j);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for GET KYC statistics\n",
                (unsigned int) response_code,
                (int) lr.hr.ec);
    break;
  }
  if (NULL != aksh->cb)
    aksh->cb (aksh->cb_cls,
              &lr);
  TALER_EXCHANGE_get_aml_kyc_statistics_cancel (aksh);
}


struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *
TALER_EXCHANGE_get_aml_kyc_statistics_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  const struct TALER_AmlOfficerPrivateKeyP *officer_priv,
  const char *names)
{
  struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *aksh;

  aksh = GNUNET_new (struct TALER_EXCHANGE_GetAmlKycStatisticsHandle);
  aksh->ctx = ctx;
  aksh->base_url = GNUNET_strdup (url);
  aksh->officer_priv = *officer_priv;
  GNUNET_CRYPTO_eddsa_key_get_public (&officer_priv->eddsa_priv,
                                      &aksh->officer_pub.eddsa_pub);
  aksh->names = GNUNET_strdup (names);
  /* Default: no start date filter, no end date filter */
  aksh->options.start_date = GNUNET_TIME_UNIT_ZERO_TS;
  aksh->options.end_date = GNUNET_TIME_UNIT_FOREVER_TS;
  return aksh;
}


enum GNUNET_GenericReturnValue
TALER_EXCHANGE_get_aml_kyc_statistics_set_options_ (
  struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *aksh,
  unsigned int num_options,
  const struct TALER_EXCHANGE_GetAmlKycStatisticsOptionValue *options)
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    const struct TALER_EXCHANGE_GetAmlKycStatisticsOptionValue *opt
      = &options[i];

    switch (opt->option)
    {
    case TALER_EXCHANGE_GET_AML_KYC_STATISTICS_OPTION_END:
      return GNUNET_OK;
    case TALER_EXCHANGE_GET_AML_KYC_STATISTICS_OPTION_START_DATE:
      aksh->options.start_date = opt->details.start_date;
      break;
    case TALER_EXCHANGE_GET_AML_KYC_STATISTICS_OPTION_END_DATE:
      aksh->options.end_date = opt->details.end_date;
      break;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_EXCHANGE_get_aml_kyc_statistics_start (
  struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *aksh,
  TALER_EXCHANGE_GetAmlKycStatisticsCallback cb,
  TALER_EXCHANGE_GET_AML_KYC_STATISTICS_RESULT_CLOSURE *cb_cls)
{
  struct TALER_AmlOfficerSignatureP officer_sig;
  CURL *eh;
  char sd_str[32];
  char ed_str[32];
  const char *sd = NULL;
  const char *ed = NULL;

  if (NULL != aksh->job)
  {
    GNUNET_break (0);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  aksh->cb = cb;
  aksh->cb_cls = cb_cls;
  TALER_officer_aml_query_sign (&aksh->officer_priv,
                                &officer_sig);
  if (! GNUNET_TIME_absolute_is_zero (aksh->options.start_date.abs_time))
  {
    unsigned long long sec;

    sec = (unsigned long long) GNUNET_TIME_timestamp_to_s (
      aksh->options.start_date);
    GNUNET_snprintf (sd_str,
                     sizeof (sd_str),
                     "%llu",
                     sec);
    sd = sd_str;
  }
  if (! GNUNET_TIME_absolute_is_never (aksh->options.end_date.abs_time))
  {
    unsigned long long sec;

    sec = (unsigned long long) GNUNET_TIME_timestamp_to_s (
      aksh->options.end_date);
    GNUNET_snprintf (ed_str,
                     sizeof (ed_str),
                     "%llu",
                     sec);
    ed = ed_str;
  }
  {
    char pub_str[sizeof (aksh->officer_pub) * 2];
    char arg_str[sizeof (aksh->officer_pub) * 2 + 32];
    char *end;

    end = GNUNET_STRINGS_data_to_string (
      &aksh->officer_pub,
      sizeof (aksh->officer_pub),
      pub_str,
      sizeof (pub_str));
    *end = '\0';
    GNUNET_snprintf (arg_str,
                     sizeof (arg_str),
                     "aml/%s/kyc-statistics/%s",
                     pub_str,
                     aksh->names);
    aksh->url = TALER_url_join (aksh->base_url,
                                arg_str,
                                "start_date",
                                sd,
                                "end_date",
                                ed,
                                NULL);
  }
  if (NULL == aksh->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  eh = TALER_EXCHANGE_curl_easy_get_ (aksh->url);
  if (NULL == eh)
  {
    GNUNET_free (aksh->url);
    aksh->url = NULL;
    return TALER_EC_GENERIC_ALLOCATION_FAILURE;
  }
  {
    struct curl_slist *job_headers = NULL;
    char *hdr;
    char sig_str[sizeof (officer_sig) * 2];
    char *end;

    end = GNUNET_STRINGS_data_to_string (
      &officer_sig,
      sizeof (officer_sig),
      sig_str,
      sizeof (sig_str));
    *end = '\0';
    GNUNET_asprintf (&hdr,
                     "%s: %s",
                     TALER_AML_OFFICER_SIGNATURE_HEADER,
                     sig_str);
    job_headers = curl_slist_append (NULL,
                                     hdr);
    GNUNET_free (hdr);
    aksh->job = GNUNET_CURL_job_add2 (aksh->ctx,
                                      eh,
                                      job_headers,
                                      &handle_get_aml_kyc_statistics_finished,
                                      aksh);
    curl_slist_free_all (job_headers);
  }
  if (NULL == aksh->job)
  {
    GNUNET_free (aksh->url);
    aksh->url = NULL;
    return TALER_EC_GENERIC_ALLOCATION_FAILURE;
  }
  return TALER_EC_NONE;
}


void
TALER_EXCHANGE_get_aml_kyc_statistics_cancel (
  struct TALER_EXCHANGE_GetAmlKycStatisticsHandle *aksh)
{
  if (NULL != aksh->job)
  {
    GNUNET_CURL_job_cancel (aksh->job);
    aksh->job = NULL;
  }
  GNUNET_free (aksh->url);
  GNUNET_free (aksh->base_url);
  GNUNET_free (aksh->names);
  GNUNET_free (aksh);
}


/* end of exchange_api_get-aml-OFFICER_PUB-kyc-statistics-NAMES.c */
